import py

def test_rename_simple(repo):

    with repo.transaction(message='create', author='test') as root:
        with root.join('test.txt').open('w') as f:
            f.write('test')

    with repo.get_default_head() as root:
        assert root.join('test.txt').exists()

    with repo.transaction(message='rename', author='test') as root:
        #XXX: check if relative names are ok for rename in the fs api
        root.join('test.txt').rename('test_renamed.txt')

    with repo.get_default_head() as root:
        assert not root.join('test.txt').exists()
        assert root.join('test_renamed.txt').exists()


@py.test.mark.xfail(reason='not implemented', run=False)
def test_rename_directory(repo):

    with repo.transaction(message='create', author='test') as root:
        dir = root.join('testdir')
        dir.mkdir()
        with dir.join('test.txt').open('w') as f:
            f.write('test')

    with repo.get_default_head() as root:
        assert root.join('testdir/test.txt').exists()

    with repo.transaction(message='rename', author='test') as root:
        #XXX: check if relative names are ok for rename in the fs api
        root.join('testdir').rename('testdir2')

    with repo.get_default_head() as root:
        assert not root.join('testdir/test.txt').exists()
        assert root.join('testdir2/test.txt').exists()




"""
    anyvc commit builder
    ~~~~~~~~~~~~~~~~~~~~

    :license: LGPL 2 or later
    :copyright: 2009 by Ronny Pfannschmidt

    Base classes for building commits in memory
"""

from os.path import join, dirname
import time as unixtime
from datetime import datetime

from anyvc.common.repository import MemoryFile

class CommitBuilder(object):
    #XXX: ugly and inflexible
    def __init__(self, repo, base_commit, time=None, local_time=True, author=None, **extra):
        self.repo = repo
        self.base_commit = base_commit
        self.extra = extra
        self.contents = {}
        self.renames = []

        if time is None:
            time = datetime.now()
        self.author = author.strip() # normalize whitespace
        self.time = time
        self.time_local = local_time

        timetuple = time.timetuple()

        self.time_unix = unixtime.mktime(timetuple)
        # timetuple[8] is the daylight saving flag
        # its -1 for normal datetimes
        # XXX: the current logic is flawed cause it only thinks about localtime
        # XXX: should it be extended to properly deal
        #      with user defined timezones via pytz/dateutil?
        #if timetuple[8] == 1 and unixtime.daylight:
        #    self.time_offset = unixtime.altzone
        #else:
        #XXX: ignores daylight saving
        self.time_offset = unixtime.timezone

    def write(self, path, content):
        self.contents[path] = content

    def read(self, path):
        if path in self.contents:
            return self.contens[path]
        for source, dest in sorted(self.renames, reverse=True):
            if path.startswith(dest):
                path = path.replace(dest, source) #XXX: incorrect, 
        return self.base_commit.file_content(path)


    def remove(self, path):
        pass

    def mkdir(self, path):
        #XXX: only relevant for vcs's that care
        pass

    def rename(self, source, dest):
        self.renames.append((source, dest))
        for name, content in self.contents.items():
            if name.startswith(source): #XXX: incorrect
                self.contents[name.replace(source, dest)] = content
                del self.contents[name]


    def commit(self):
        raise NotImplementedError

    def __enter__(self): 
        return RevisionBuilderPath(self.base_commit, "/", self)

    def __exit__(self, etype,  eval, tb):
        if etype is None: 
            self.commit()


class RevisionBuilderPath( object):
    def __init__( self, commit, path, builder):
        self.commit = commit
        self.path = path
        self.builder = builder


    def mkdir(self):
        self.builder.mkdir(self.path)

    def rename(self , new_name):
        new = self.parent().join(new_name)
        assert self.path != '/' and new_name != '/'
        self.builder.rename(self.path, new.path)
        #XXX: shoould self.path change?

    def parent(self):
        return RevisionBuilderPath(self.commit, dirname(self.path), self.builder)

    def join(self,  path):
        return RevisionBuilderPath(self.commit, join(self.path, path), self.builder)

    def open(self,  mode='r'):
        #implement in terms of read/write
        if mode ==  'r':
            raise NotImplementedError
        elif mode == 'w':
            #XXX: test and implement all flavors of reopening
            return FileBuilder(path=self)

    def write(self, data):
        self.builder.write(self.path, data)

    def read(self):
        return self.builder.read(self.path)

    def check(self):
        try:
            self.read() #XXX: less expensive?
            return True
        except IOError:
            return False


class FileBuilder(MemoryFile):
    def __exit__(self, et, ev, tb):
        # subvertpy file data transfer doesn't seek back
        self.path.write(self.getvalue())





def test_diff_create_simple(repo):

    with repo.transaction(author='test', message='test') as root:
        with root.join('test.txt').open('w') as f:
            f.write('test')

    head = repo.get_default_head()
    diff = head.get_parent_diff()
    assert '+test' in diff



# -*- coding: utf-8 -*-
# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
"""
    weird base classes

    :copyright: 2008 Ronny Pfannschmidt
    :license: LGPL2 or later
"""
from py.path import local
from os.path import join

from os.path import dirname, basename, join, normpath

class StatedPath(object):
    """
    stores status informations about files

    >>> StatedPath('a.txt')
    <normal 'a.txt'>
    >>> StatedPath('a.txt', 'changed')
    <changed 'a.txt'>

    """

    def __init__(self, name, state='normal', base=None):
        self.relpath = normpath(name)
        self.path = dirname(name)
        self.name = basename(name)
        self.base = base
        self.state = intern(state)
        if base is not None:
            self.abspath = local(base).join(name)
        else:
            self.abspath = None

    def __repr__(self):
        return '<%s %r>'%(
                self.state,
                self.relpath,
                )

    def __str__(self):
        return self.relpath


def find_basepath(act_path, check):
    """
    :param act_path: starting dir
    :param check: the test

    a helper function walks the directory tree up till it finds dir
    for wich the the check is true
    """

    act_path = local(act_path)
    # this logic kind of fails for svn, but who cares
    for part in act_path.parts(reverse=True):
        if check(part):
            return part

class WorkDir(object):
    """
    Basic Workdir API

    :param path: base path
    :param create: 
    """

    def __init__(self, path, create=False, source=None):
        self.path = path
        if create:
            if source:
                self.create_from(source)
            else:
                self.create()
        else:
            pass #XXX


    def process_paths(self, paths):
        """
        preprocess given paths
        """
        return list(paths)

    def status(self, paths=(), recursive=True):
        """
        :param path: the filenames
        :type path: sequence of string
        :param recursive: proceed recursive for directories
        :type recursive: bool

        yield a list of Path instances tagged with status informations
        """
        raise NotImplementedError

    def diff(self, paths=()):
        """
        given a list of paths it will return a diff
        """
        raise NotImplementedError

    def update(self, paths=(), revision=None):
        """
        :param revision: the target revision
                         may not actually work for
                         vcs's with tricky workdir revision setups

        updates the workdir to either the closest head or or the given revision
        """
        raise NotImplementedError

    def commit(self, paths=(), message=None, user=None):
        """
        :param path: the paths
        :param message: the commit message
        :param user: optional author name

        commits the given paths/files with the given commit message and author
        """

        raise NotImplementedError

    def revert(self, paths=None, missing=False):
        raise NotImplementedError

    def add(self, paths=None, recursive=False):
        raise NotImplementedError

    def remove(self, paths=None, execute=False, recursive=False):
        raise NotImplementedError

    def rename(self, source=None, target=None):
        raise NotImplementedError


class WorkDirWithParser(WorkDir):
    """
    extension of the workdir class to support
    parsing needs
    """


    def parse_status_items(self, items, cache):
        """
        default implementation

        for each `item` in `items` invoke::

            self.parse_status_item(item, cache)

        .. note::
            a more complex parser might need to overwrite
        """
        for item in items:
            rv = self.parse_status_item(item, cache)

            if rv is not None:
                state, name = rv
                #XXX: here renames get turned into ugly add/remove pairs
                if state is None:
                    old, new = name
                    yield StatedPath(old, 'removed', self.path)
                    yield StatedPath(new, 'added', self.path)
                else:
                    yield StatedPath(name, state, self.path)


    def parse_status_item(self, item, cache):
        """
        parse a single status item
        meant to be overridden
        """
        raise NotImplementedError

    def parse_cache_items(self, items):
        """
        parses vcs specific cache items to a list of (name, state) tuples
        """
        return []

    def cache_impl(self, paths=False, recursive=False):
        """
        creates a list of vcs specific cache items
        only necessary by messed up vcs's

        in case of doubt - dont touch ^^
        """
        return []

    def status_impl(self, paths=False, recursive=False):
        """
        yield a list of vcs specific listing items
        """
        raise NotImplementedError

    def cache(self, paths=(), recursive=False):
        """
        return a mapping of name to cached states
        only necessary for messed up vcs's
        """
        return dict(
                self.parse_cache_items(
                self.cache_impl(
                    paths = paths,
                    recursive=recursive
                    )))

    def status(self, paths=(), recursive=True):
        """
        yield a list of Path instances tagged with status informations
        """
        cache = self.cache(paths = paths,recursive=recursive)
        return self.parse_status_items(
                self.status_impl(
                    paths = paths,
                    recursive=recursive,
                    ), cache)


import re

from subprocess import Popen, PIPE, STDOUT, call
import os, os.path


def relative_to(base_path):
    """
    will turn absolute paths to paths relative to the base_path

    .. warning:
        will only work on paths below the base_path
        other paths will be unchanged
    """
    base_path = local(base_path)
    def process_path(child_path):
        child_path = local(child_path)
        if child_path.relto(base_path):
            return base_path.bestrelpath(child_path)
        else:
            return child_path
    return process_path


class CommandBased(WorkDirWithParser):
    """
    common code + default method implementations
    for subprocess based vcs's
    """
    #TODO: set up the missing actions





    def execute_command(self, args, result_type=str, **kw):
        if not args:
            raise ValueError('need a valid command')
        ret = Popen(
                [self.cmd] + [str(x) for x in args], # str is for py.path
                stdout=PIPE,
                stderr=STDOUT,
                cwd=self.path.strpath,
                close_fds=True,
                env=dict(os.environ, LANG='C',LANGUAGE='C', LC_All='C'),
                )
        if result_type is str:
            return ret.communicate()[0]
        elif result_type is iter:
            return ret.stdout
        elif result_type is file:
            return ret.stdout

    def get_commit_args(self, message, paths=(), **kw):
        """
        creates a argument list for commiting

        :param message: the commit message
        :param paths: the paths to commit
        """
        return ['commit','-m', message] + self.process_paths(paths)

    def get_diff_args(self, paths=(), **kw):
        return ['diff'] + self.process_paths(paths)

    def get_update_args(self, revision=None, **kw):
        if revision:
            return ['update', '-r', revision]
        else:
            return ['update']

    def get_add_args(self, paths=(), recursive=False, **kw):
        return ['add'] + self.process_paths(paths)

    def get_remove_args(self, paths=(), recursive=False, execute=False, **kw):
        return ['remove'] +  self.process_paths(paths)

    def get_revert_args(self, paths=(), recursive=False, **kw):
        return ['revert'] + self.process_paths(paths)

    def get_status_args(self,**kw):
        raise NotImplementedError("%s doesn't implement status"%self.__class__.__name__)

    def commit(self, **kw):
        args = self.get_commit_args(**kw)
        return self.execute_command(args, **kw)

    def diff(self, **kw):
        args = self.get_diff_args(**kw)
        return self.execute_command(args, **kw)

    def update(self, **kw):
        args = self.get_update_args(**kw)
        return self.execute_command(args, **kw)

    def add(self, **kw):
        args = self.get_add_args(**kw)
        return self.execute_command(args, **kw)

    def remove(self, **kw):
        args = self.get_remove_args(**kw)
        return self.execute_command(args, **kw)

    def rename(self, source, target, **kw):
        args = self.get_rename_args(source, target)
        return self.execute_command(args, **kw)

    def revert(self, **kw):
        args = self.get_revert_args(**kw)
        return self.execute_command(args, **kw)

    def status_impl(self, **kw):
        """
        the default implementation is only cappable of
        recursive operation on the complete workdir

        rcs-specific implementations might support
        non-recursive and path-specific listing
        """
        args = self.get_status_args(**kw)
        return self.execute_command(args, result_type=iter, **kw)

    def cache_impl(self, recursive, **kw):
        """
        implement caching by running a subprocess
        only runs caching if it knows, how
        """
        args = self.get_cache_args(**kw)
        if args:
            return self.execute_command(args, result_type=iter, **kw)
        else:
            return []

    def get_cache_args(self, **kw):
        return None








# copyright 2008 by Ronny Pfannschmidt
from anyvc.common.workdir import StatedPath
from py.path import local

def test_no_abspath():
    assert StatedPath('a').abspath is None

def test_base_to_abspath():
    assert StatedPath('a', base=local('b')).abspath == local('b/a')

def test_repr():
    assert repr(StatedPath('a')) == "<normal 'a'>"
    assert repr(StatedPath('./a')) == "<normal 'a'>"





def test_find_repositories(mgr):
    from anyvc import repository
    names = ['fun','test','oops',]
    repos = [mgr.make_repo(name) for name in names]
    print mgr.base.listdir()

    found_repos = list(repository.find(mgr.base, backends=[mgr.vc]))
    for repo in found_repos:
        print repo
    assert len(found_repos) == len(names)




"""
    Anyvc Mercurial repository support
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :license: LGPl 2 or later
    :copyright:
        * 2008 by Ronny Pfannschmidt <Ronny.Pfannschmidt@gmx.de>
"""

from anyvc.common.repository import Repository, Revision, join
from anyvc.common.commit_builder import CommitBuilder
from .workdir import grab_output
from datetime import datetime
from mercurial import commands, localrepo, ui, context
from mercurial import error
from ..exc import NotFoundError


class MercurialRevision(Revision):
    def __init__(self, repo, rev):
        self.repo, self.rev = repo, rev
    
    @property
    def id(self):
        return self.rev.hex()

    @property
    def author(self):
        return self.rev.user()

    @property
    def time(self):
        return datetime.fromtimestamp(self.rev.date()[0])

    @property
    def parents(self):
        return [MercurialRevision(self.repo, rev) for rev in self.rev.parents() if rev]


    @property
    def message(self):
        return self.rev.description()

    def file_content(self, path):
        try:
            return self.rev[path].data()
        except LookupError:
            raise IOError('%r not found'%path)

    def get_changed_files(self):
        return self.rev.files()

    def isdir(self, path):
        entries = list(self.rev)
        if path[-1]!='/':
            path+='/'
        from bisect import bisect_right
        entry = entries[bisect_right(entries, path)]
        return path!=entry and path.startswith(entry)

    def isfile(self, path):
        return path in self.rev

    def exists(self, path):
        return self.isfile(path) or self.isdir(path)

class MercurialCommitBuilder(CommitBuilder):
    def commit(self):
        repo = self.repo.repo
        def get_file(repo, ctx, path):
            #XXX: copy sources
            #XXX: renames
            #XXX: deletes
            if path in rn and path not in self.contents:
                raise IOError()
            if path in rrn:
                assert base is not None
                parent = self.base_commit.rev[rrn[path]]
                if path in self.contents:
                    data = self.contents[path]
                else:
                    data = parent.data()
                copyed = rrn[path]
            else:
                data = self.contents[path]
                copyed = False

            islink = False #XXX
            isexec = False #XXX


            return context.memfilectx(path, data, islink, isexec, copyed)

        rn = dict((k, v) for k, v in self.renames if self.base_commit.exists(k))
        rrn = dict(reversed(x) for x in self.renames)
        #XXX: directory renames

        files = set(self.contents)
        files.update(rn.keys())
        files.update(rn.values())

        if self.base_commit is not None:
            base = self.base_commit.rev.node()
        else:
            base = None
        ctx = context.memctx(
                repo,
                [base, None],
                self.extra['message'],
                sorted(files),
                get_file,
                user=self.author,
                date="%(time_unix)d %(time_offset)s"%self.__dict__,
                )
        repo.commitctx(ctx)


class MercurialRepository(Repository):
    CommitBuilder = MercurialCommitBuilder

    def __init__(self, path=None, workdir=None, create=False):
        self.path = path
        self.workdir = workdir
        #XXX: path only handling
        if workdir is not None:
            self.repo = workdir.repo
            self.ui = self.repo.ui

        elif path is not None:
            try:
                repo = localrepo.localrepository(ui.ui(), str(path), create=create)
            except error.RepoError:
                raise NotFoundError('mercurial', path)

            self.ui = repo.ui
            self.repo = repo

    def invalidate_cache(self):
        self.repo.invalidate()

    @grab_output
    def push(self, dest=None, rev=None):
        self.invalidate_cache()
        commands.push(self.ui, self.repo, dest, rev=rev)

    @grab_output
    def pull(self, source="default", rev=None):
        self.invalidate_cache()
        commands.pull(self.ui, self.repo, source, rev=rev)

    def __len__(self):
        return len(self.repo)

    def get_default_head(self):
        return self['tip']

    def __getitem__(self, id):
        self.invalidate_cache()
        return MercurialRevision(self, self.repo[id])



import py
from tests.helpers import  VcsMan
import os

from anyvc import metadata

pytest_plugins = "doctest"

test_in_interpreters = 'python2', 'python3', 'jython', 'pypy'

test_on = {
    '%s': None,
    'remote/%s': 'popen//python=python2',
}

def pytest_addoption(parser):
    g = parser.getgroup('anyvc')
    g.addoption("--local-remoting", action="store_true", default=False,
               help='test via execnet remoting')
    g.addoption("--no-direct-api", action="store_true", default=False,
                help='don\'t test the direct api')
    g.addoption("--vcs", action='store', default=None)

def pytest_configure(config):
    if not config.getvalue('local_remoting'):
        for key in list(test_on):
            if '/' in key:
                del test_on[key]
    if config.getvalue('no_direct_api'):
        for key in list(test_on):
            if '/' not in key:
                del test_on[key]

    assert test_on,  'you shouldnt disable all test variations'

    vcs = config.getvalue('vcs')
    if vcs is None:
        return
    vcs = vcs.split('-')[0]
    if vcs not in metadata.backends:
        if vcs in metadata.aliases:
            vcs = metadata.aliases[vcs]
            config.option.vcs = vcs
        else:
            raise KeyError(vcs, '%r not found' % vcs)


    os.environ['BZR_EMAIL'] = 'Test <test@example.com>'

funcarg_names = set('mgr repo wd backend'.split())

def pytest_generate_tests(metafunc):
    if not funcarg_names.intersection(metafunc.funcargnames):
        return
    for name in metadata.backends:
        wanted = metafunc.config.getvalue('vcs')
        if wanted is not None and name!=wanted:
            continue
        for id, spec in test_on.items():
            metafunc.addcall(id=id%name, param=(name, spec))


def pytest_funcarg__backend(request):
    """
    create a cached backend instance that is used the whole session
    makes instanciating backend cheap
    """
    vc, spec = request.param
    return request.cached_setup(
            lambda: metadata.get_backend(vc, spec),
            extrakey=request.param,
            scope='session')


def pytest_funcarg__mgr(request):
    """
    create a preconfigured :class:`tests.helplers.VcsMan` instance
    pass the currently tested backend 
    and create a tmpdir for the vcs/test combination

    auto-check for the vcs features and skip if necessary
    """
    vc, spec = request.param
    r = spec or 'local'
    vcdir = request.config.ensuretemp('%s_%s'%(vc, r) )
    testdir = vcdir.mkdir(request.function.__name__)
    backend = request.getfuncargvalue('backend')

    required_features = getattr(request.function, 'feature', None)

    if required_features:
        required_features = set(required_features.args)
        difference = required_features.difference(backend.features)
        print required_features
        if difference:
            py.test.skip('%s lacks features %r' % (vc, sorted(difference)))

    return VcsMan(vc, testdir, spec, backend)

def pytest_funcarg__repo(request):
    """
    create a repo below mgf called 'repo'
    """
    return request.getfuncargvalue('mgr').make_repo('repo')

def pytest_funcarg__wd(request):
    """
    create a workdir below mgr called 'wd'
    if the feature "wd:heavy" is not supported use repo as help
    """
    mgr = request.getfuncargvalue('mgr')
    if 'wd:heavy' not in mgr.backend.features:
        repo = request.getfuncargvalue('repo')
        wd = mgr.create_wd('wd', repo)
    else:
        wd = mgr.create_wd('wd')

    fp = request.function
    if hasattr(fp, 'files'):
        files = fp.files.args[0]
        wd.put_files(files)
        assert wd.has_files(*files)
        if  hasattr(fp, 'commit'):
            wd.add(paths=list(files))
            wd.commit(message='initial commit')
    return wd


def pytest_collect_directory(path, parent):
    for compiled_module in path.listdir("*.pyc"):
        if not compiled_module.new(ext=".py").check():
            compiled_module.remove()




import py.test
from datetime import datetime


def test_repo_create(repo):
    default_branch = repo.prepare_default_structure()
    assert len(repo) in (0,1)

@py.test.mark.feature('dvcs')
def test_repo_default_head(wd):
    repo = wd.repository
    wd.put_files({'test.py': "import sys\nprint sys.platform" })
    wd.add(paths=['test.py'])
    wd.commit(message="test commit")
    for i, message in enumerate(["test commit", "test commit\n", "test\nabc"]):
        wd.put_files({'test.py':'print %s'%i})
        wd.commit(message=message)
        head = repo.get_default_head()
        print repr(head.message), repr(message)
        #XXX: how to propperly normalize them
        assert head.message.strip()==message.strip()




from execnet import makegateway
import pickle
from anyvc.remote.object import RemoteCaller

def test_remote_object():
    gw = makegateway('popen')
    channel = gw.remote_exec("""
    from anyvc.remote.object import RemoteHandler

    class TestHandler(RemoteHandler):
        def test(self, **kw):
            return kw
    newchan = channel.gateway.newchannel()
    handler = TestHandler(newchan)
    channel.send(newchan)
    """)
    caller = RemoteCaller(channel.receive())
    result = caller.test(a=1)
    assert result == {'a':1}


class TestChannel(object):

    def isclosed(self):
        return hasattr(self, 'next')

    def send(self, val):
        print val
        self.next = val

    def receive(self):
        result = self.next
        del self.next #XXX: crude hack
        return result

def test_caller():

    test_handler = RemoteCaller(TestChannel())
    method, args, kwargs = test_handler.test(a=1)

    assert method == 'test'
    assert kwargs == {'a': 1}
    assert args == ()



import py
from datetime import datetime


def test_build_first_commit(repo):
    with repo.transaction(message='initial', author='test') as root:
        with root.join('test.txt').open('w') as f:
            f.write("text")

    head = repo.get_default_head()
    with head as root:
        with root.join("test.txt").open() as f:
            content = f.read()
            assert content == 'text'

    if not isinstance(head.id, int):
        hre = py.std.re.compile('\w+')
        assert hre.match(head.id), 'rev id is messed goo %r'%head.id

def test_generate_commit_chain(repo):
    for i in range(1,11):
        with repo.transaction(
                message='test%s'%i,
                author='test') as root:
            with root.join('test.txt').open('w') as f:
                f.write("test%s"%i)

    assert len(repo) == 10

    head = repo.get_default_head()

    revs = [head]
    rev = head

    while rev.parents:
        rev = rev.parents[0]
        revs.append(rev)

    assert len(revs) == 10

    for i, rev in enumerate(reversed(revs)):
        with rev as root:
            with root.join('test.txt').open() as f:
                data = f.read()
                assert data == 'test%s'%(i+1)


def test_create_commit_at_time(mgr, repo):
    if mgr.vc == 'subversion':
        py.test.skip('currently no support for setting the commit time on svn')

    time = datetime(2000, 1, 1, 10, 0, 0)

    with repo.transaction(
            message='test',
            author='test',#XXX: author should be optional
            time=datetime(2000, 1, 1, 10, 0, 0)) as root:
        with root.join('test.txt').open('w') as f:
                f.write('test')

    head = repo.get_default_head()

    print repr(head.id)
    print head.time
    assert head.time == time


def test_create_commit_with_author(mgr):
    if mgr.vc == 'subversion':
        py.test.skip('currently no support for setting the commit author on svn')

    repo = mgr.make_repo('repo')

    with repo.transaction(
            message='test',
            author='test author ', #with whitespace
            ) as root:
        with root.join('test.txt').open('w') as f:
                f.write('test')

    head = repo.get_default_head()
    print repr(head.author)
    assert head.author == 'test author' #whitespace gone




import py
import execnet


@py.test.mark.files({'setup.py':'pass'})
@py.test.mark.commited
def test_diff(wd):
    gw = execnet.makegateway()
    def remote(channel, cwd):
        import sys
        import os
        import anyvc
        #XXX to cause the imports in case of relatives
        #    this should get fixed in execnet
        anyvc.workdir.open(cwd)
        import anyvc.client
        os.chdir(cwd)
        channel.send(anyvc.client.main(['vc','diff']))

    ch = gw.remote_exec(remote, cwd = wd.path.strpath)
    ch.receive()



from anyvc.backend import Backend
from anyvc.remote.object import RemoteHandler
import time
import datetime

def start_controller(channel):
    vcs = channel.receive()
    backend_module = channel.receive()
    try:
        backend = Backend(vcs, backend_module)
        backend.module # cause that one is lazy-loaded
    except ImportError:
        channel.send(None) # magic value for 'i dont have it'
        return

    workchan = channel.gateway.newchannel()
    SlaveBackend(workchan, backend)
    channel.send(workchan)


class SlaveBackend(RemoteHandler):
    def __init__(self, channel, backend):
        RemoteHandler.__init__(self, channel)
        self.backend = backend


    def open_repo(self, *k,**kw):
        repo = self.backend.Repository(*k, **kw)
        channel = self.newchannel()
        RepositoryHandler(channel, repo)
        return channel

    def is_repository(self, path):
        return self.backend.is_repository(path)

    def is_workdir(self, path):
        return self.backend.is_workdir(path)

    def open_workdir(self, *k, **kw):
        workdir = self.backend.Workdir(*k, **kw)
        channel = self.newchannel()
        WorkdirHandler(channel, workdir)
        return channel

    def features(self):
        return self.backend.features

class RepositoryHandler(RemoteHandler):
    def __init__(self, channel, repo):
        RemoteHandler.__init__(self, channel)
        self.repo = repo

    def push(self):
        self.repo.push()

    def get_default_head(self):
        return self.repo.get_default_head().id

    def commit_message(self, id):
        return self.repo[id].message

    def path(self):
        return str(self.repo.path)

    def commit_diff(self, id):
        return self.repo[id].get_parent_diff()

    def commit_exists(self, id, path):
        return self.repo[id].exists(path)

    def commit_file_content(self, id, path):
        try:
            return self.repo[id].file_content(path)
        except IOError:
            return None

    def commit_parents(self, id):
        return [p.id for p in self.repo[id].parents]

    def commit_time(self, id):
        return time.mktime(self.repo[id].time.timetuple())

    def commit_author(self, id):
        return self.repo[id].author

    def prepare_default_structure(self):
        self.repo.prepare_default_structure()

    def count_revisions(self):
        return len(self.repo)

    def transaction(self, **kw):
        time = kw.get('time')
        if time is not None:
            kw['time'] = datetime.datetime.fromtimestamp(time)
        transaction = self.repo.transaction(**kw)
        channel = self.newchannel()
        TransactionHandler(channel, transaction, self.repo)
        return channel


class WorkdirHandler(RemoteHandler):
    def __init__(self, channel, workdir):
        RemoteHandler.__init__(self, channel)
        self.workdir = workdir

    def get_path(self):
        return str(self.workdir.path)

    def add(self, **kw):
        return self.workdir.add(**kw)

    def status(self, **kw):
        return [(item.relpath, item.base.strpath, item.state)
                for item in self.workdir.status(**kw)]

    def commit(self, **kw):
        return self.workdir.commit(**kw)

    def diff(self, **kw):
        return self.workdir.diff(**kw)

    def remove(self, **kw):
        return self.workdir.remove(**kw)

    def revert(self, **kw):
        return self.workdir.revert(**kw)

    def rename(self, **kw):
        return self.workdir.rename(**kw)

    def get_local_repo(self):
        #XXX: this one shouldnt be
        repo = self.workdir.repository
        if repo is not None:
            channel = self.newchannel()
            RepositoryHandler(channel, repo)
            return channel


class TransactionHandler(RemoteHandler):
    def __init__(self, channel, transaction, repo):
        RemoteHandler.__init__(self, channel)
        self.transaction = transaction
        self.repo = repo
        self.transaction.__enter__()

    def write(self, path, data):
        fb = self.transaction.write(path, data)

    def commit(self):
        self.transaction.__exit__(None, None, None)

    def rename(self, source, dest):
        self.transaction.rename(source, dest)




"""
    anyvc
    ~~~~~~

    pythonic vcs abstractions

    :license: LPL2
    :copyright: Ronny Pfannschmidt and others
"""


import apipkg
apipkg.initpkg(__name__,{
    'workdir':{
        'clone':'anyvc._workdir:clone',
        'checkout': 'anyvc._workdir:checkout',
        'open':'anyvc._workdir:open',
        },
    })



"""
    git backend
    ----------------

    :copyright: 2007-2008 Ronny Pfannschmidt
    :license: LGPL2 or later
"""



from anyvc.common.workdir import CommandBased, relative_to
import re
from subprocess import call

class Git(CommandBased):
    """
    fooked liek hell
    git wants 3 subprocess calls to get all needed infos
    data might be wrong in coner cases
    """
    cmd = 'git'

    @property
    def repository(self):
        from .repo import GitRepository
        return GitRepository(workdir=self)

    def setup(self):
        pass #XXX: actually unneded?

    def create_from(self, source):
        call(['git', 'clone', str(source), self.path.strpath])

    def create(self):
        call(['git', 'init', self.path.strpath])

    def get_diff_args(self, paths=(), **kw):
        return ['diff', '--no-color'] + self.process_paths(paths)

    def get_commit_args(self, message, paths=(), **kw):
        if paths:
            # commit only for the supplied paths
            return ['commit', '-m', message, '--'] + list(self.process_paths(paths))
        else:
            # commit all found changes
            # this also adds all files not tracked and not in gitignore
            # this also commits deletes ?!
            return ['commit', '-a', '-m', message]

    def get_revert_args(self, paths=(), recursive=False, **kw):
        return ['checkout','HEAD'] + self.process_paths(paths)

    def get_remove_args(self, paths=(), recursive=False, execute=False, **kw):
        return ['rm'] +  self.process_paths(paths)

    def get_rename_args(self, source, target):
        return ['mv', source, target]

    def parse_status_item(self, item, cache):
        return item

    def status_impl(self, *k, **kw):
        #XXX: OMG HELLISH FRAGILE SHIT
        if self.execute_command(['branch']).strip():
            tree = set(self.execute_command([
            'ls-tree', '-r', '--name-only', 'HEAD'
            ]).splitlines())
        else:
            tree = set()

        def ls_files(args):
            files = self.execute_command([
                    'ls-files' ] + [
                    '-%s' % c for c in args
                ]).splitlines()

            d = dict()
            for item in files:
                state, name = item.split(" ", 1)
                d.setdefault(name, set()).add(state)
            return dict(reversed(x.split(" ", 1))
                        for x in files)

        wd = ls_files("tdmo")
        index = ls_files("tcdo")
        all = sorted(tree | set(wd) | set(index))

        for name in all:
            it = name in tree
            w = wd.get(name, [])
            i = index.get(name, [])
            #XXX: factor into parse_status_item
            if '?' in w:
                yield 'unknown', name
            elif 'H' in i and not w and not it:
                yield 'added', name
            elif 'H' in i and not w and it:
                yield 'clean', name
            elif not w and not i and it:
                yield 'removed', name

            elif 'C' in w and 'R' in i:
                yield 'missing', name
            elif 'C' in w and 'H' in i:
                yield 'modified', name





"""
    Anyvc repo support

    :license: LGPL 2 or later
    :copyright: 2009 by Ronny Pfannschmidt
"""
import os
from anyvc.exc import NotFoundError
from py.path import local

def open(path, backends=None):
    """
    :param backends: optional list of backends to try

    open a repository backend at the given path
    """
    from anyvc.metadata import get_backends
    for backend in get_backends(backends):
        if backend.is_repository(path):
            #XXX: add metadata about the worktree base, use it
            return backend.Repository(path)

def find(root, backends=None):
    """
    :param root: the search root
    :type  root: py.path.local or path string

    find all repositories below `root`
    """
    start = local(root)
    backend = open(root, backends=backends)
    if backend is not None:
        yield backend
    else:
        for subdir in root.listdir(lambda p: p.check(dir=1)):
            for item in find(subdir, backends):
                yield item




import os

from anyvc.common.workdir import CommandBased, relative_to, WorkDirWithParser
from subprocess import call
#from subvertpy import client, wc, ra, NODE_DIR

#from subvertpy.ra import RemoteAccess, Auth, get_username_provider, SubversionException

class SubVersion(CommandBased):
    #XXX: disabled
    cmd = "svn"
    detect_subdir = ".svn/props"
    repository = None # no local repo


    def create(self):
        raise NotImplementedError('no havy workdirs')

    def get_status_args(self, recursive, paths, **kw):
        #TODO: figure a good way to deal with changes in external
        # (maybe use the svn python api to do that)
        ret = ["st", "--no-ignore", "--ignore-externals", "--verbose"]
        if not recursive:
            ret.append("--non-recursive")
        return ret + list(paths)


    def create_from(self, source):
        call(['svn', 'co', 'file://%s' % source, self.path.strpath])

    state_map = {
            "?": 'unknown',
            "A": 'added',
            " ": 'clean',
            "!": 'missing',
            "I": 'ignored',
            "M": 'modified',
            "D": 'removed',
            "C": 'conflict',
            'X': 'clean',
            'R': 'modified',
            '~': 'clean',
            }

    def get_add_args(self, paths, **kw):
        # svn add doesnt add parent dirs by default
        return ['add', '--parents'] + paths

    def get_diff_args(self, paths=(), **kw):
        return ['diff', '--diff-cmd', 'diff'] + list(paths)

    def get_rename_args(self, source, target):
        return ['move', source, target]

    def parse_status_item(self, item, cache):
        if item[0:4] == 'svn:':
            # ignore all svn error messages
            return None
        state = item[0]
        file = item.split()[-1]
        if file == '.':
            # this is the path of the repo
            # normalize to ''
            file = ''
        #TODO: handle paths with whitespace if ppl fall in that one
        return self.state_map[state], file





class BrokenSubversion(object):  # workdir with parser
    detect_subdir= '.svn/entries'
    repository = None # no local repo

    def create_from(self, source):
        #XXX: omg what a fucked up mess
        r = ra.RemoteAccess('file://' + source)
        rev = r.get_latest_revnum()
        print rev
        import os
        #XXX: wth are we doing here
        os.mkdir(self.path)
        c = client.Client(auth=Auth([get_username_provider()]))
        c.checkout('file://' + source, self.path, rev)

    def add(self, paths=None, recursive=False):
        #XXX: recursive
        import os
        #XXX: hacl
        print paths
        w = wc.WorkingCopy(
                path=self.path,
                write_lock=True,
                associated=None)
        for path in paths:
            segments = path.split(os.path.sep)
            p = self.path
            for segment in segments:
                p = os.path.join(p, segment)
                print p, w.add(path=p)

        w.close()

    def commit(self, paths=None, message=None, user=None):
        if paths:
            targets = [os.path.join(self.path, path) for path in paths]
        else:
            targets = [self.path]
        c = client.Client(auth=Auth([get_username_provider()]))

        def m(items):
            print items
            print message
            return message #XXX: encoding
        c.log_msg_func = m
        c.commit(targets=targets, recurse=True)

    def walk_status(self, parent, basename):
        basepath = os.path.join(self.path, basename)
        w = wc.WorkingCopy(parent, basepath)
        e = w.entries_read(True)
        print e.keys()
        for name, entry in e.items():
            if not name: #ignore root for now
                continue 
            yield os.path.join(basename, name), entry
            if entry.kind == NODE_DIR:
                for item in self.walk_status(w, os.path.join(basename, name)):
                    yield item
        others = os.listdir(basepath)

        for item in others:
            if item=='.svn' or item in e:
                continue
            print item
            yield os.path.join(basename, item), None #unknown


    def status_impl(self,paths=(), recursive=True):
        return self.walk_status(None, '') #xxx: set limits?

    def parse_status_item(self, item, cache):
        name, e = item
        if e is None:
            return 'unknown', name
        print e.kind, e.schedule, repr(e.name)
        map = {
            wc.SCHEDULE_ADD: 'added',
            wc.SCHEDULE_DELETE: 'removed',
            wc.SCHEDULE_NORMAL: 'clean',
        }
        state = map[e.schedule]
        full_path = os.path.join(self.path, name)
        if not os.path.exists(full_path) and state!='removed':
            return 'missing', name
        print state, name
        if state=='clean':
            import hashlib
            with open(os.path.join(self.path, name)) as f:
                if hashlib.md5(f.read()).hexdigest()!=e.checksum:
                    return 'modified', name

        return state, name




import py
from anyvc.remote import RemoteBackend
from anyvc.metadata import backends

def test_end_popen_backend(backend):
    backend = RemoteBackend(backend.name, backends[backend.name], 'popen')
    assert backend.active
    backend.stop()
    assert not backend.active

def test_missing_backend_failure(monkeypatch):
    print backends
    py.test.raises(ImportError, RemoteBackend, 'testvc', '_missing_', 'popen')



"""
    anyvc.metadata
    ~~~~~~~~~~~~~~

    some basic metadata about vcs states and other fun

    .. warning::

      this module is subject to huge changes
"""

def _(str):
    return str #XXX: gettext

state_descriptions = dict(
    #XXX: finish, validate
    unknown = _("not known to the vcs"),
    ignored = _("ignored by the vcs"),
    added = _("added"),
    clean = _("known by the vcs and unchanged"),
    modified =_("changed in the workdir"),
    missing = _("removed from the workdir, still recorded"),
    removed = _("removed by deletion or renaming"),
    conflicte = _("merge conflict")
)

aliases = {
    'svn': 'subversion',
    'bzr': 'bazaar',
    'hg': 'mercurial',
}


# known implementations

backends = {
    'mercurial': 'anyvc.mercurial',
    'bazaar': 'anyvc.bazaar',
    'git': 'anyvc.git',
    'subversion': 'anyvc.subversion'
}

def get_backends(limit_to=None):
    """
    :param limit_to: optional list of backends to try, will use all if use is none
    :type limit_to: list of string or None

    a generator over all known backends
    """
    use = limit_to or backends
    for backend in use:
        try:
            yield get_backend(backend)
        except ImportError, e:
            pass


def get_backend(vcs, use_remote=False):
    module = backends[vcs]
    if use_remote is True:
        use_remote = 'popen'
    if use_remote:
        from anyvc.remote import RemoteBackend
        return RemoteBackend(vcs, module, use_remote)
    else:
        from anyvc.backend import Backend
        return Backend(vcs, module)

def get_wd_impl(vcs):
    return get_backend(vcs).Workdir


def get_repo_impl(vcs):
    return get_backend(vcs).Repository



"""
    anyvc.remote
    ~~~~~~~~~~~~

    run packends in different processes and on different computers
    also circumvent the gpl


    :license: lgpl2
    :copyright: 2009 by Ronny Pfannschmidt

"""
from execnet import makegateway
from os.path import join

from anyvc.exc import NotFoundError
from anyvc.util import cachedproperty
from .object import RemoteCaller
from anyvc.common.commit_builder import CommitBuilder, FileBuilder, RevisionBuilderPath
from anyvc.common.repository import MemoryFile
from anyvc.metadata import backends
import time
import datetime
from py.path import local

class RemoteCommit(object):
    def __init__(self, repo, id):
        self.repo = repo
        self.id = id

    def get_parent_diff(self):
        return self.repo.commit_diff(self.id)

    def exists(self, path):
        return self.repo.commit_exists(self.id, path)

    def file_content(self, path):
        data = self.repo.commit_file_content(self.id, path)
        if data is None:
            raise IOError('%r not found'%path)
        return data

    @cachedproperty
    def parents(self):
        return [RemoteCommit(self.repo, id)
                for id in self.repo.commit_parents(self.id)]

    @cachedproperty
    def author(self):
        return self.repo.commit_author(self.id)

    @cachedproperty
    def message(self):
        return self.repo.commit_message(self.id)

    def __enter__(self):
        from anyvc.common.repository import RevisionView
        return RevisionView(self, '')

    @cachedproperty
    def time(self):
        return datetime.datetime.fromtimestamp(self.repo.commit_time(self.id))

    def __exit__(self, et, ev, tb):
        pass


class RemoteRepository(RemoteCaller):
    
    @cachedproperty
    def path(self):
        return local(self._call_remote('path'))

    def get_commit(self, id):
        return RemoteCommit(self, id)

    def get_default_head(self):
        id = self._call_remote('get_default_head')
        return self.get_commit(id)

    def transaction(self, *k, **kw):
        t = kw.get('time')
        if isinstance(t, datetime.datetime): #XXX: fragile
            t = time.mktime(t.timetuple())
            kw['time'] = t
        channel = self._call_remote('transaction', *k, **kw)
        return RemoteTransaction(channel)

    def __len__(self):
        return self._call_remote('count_revisions')


class RemoteWorkdir(RemoteCaller):

    def status(self, **kw):
        from anyvc.common.workdir import StatedPath
        items = self._call_remote('status', **kw)
        for path, base, state in items:
            yield StatedPath(path, state, base)

    @property
    def repository(self):
        #XXX: this one shouldnt be
        channel = self._call_remote('get_local_repo')
        if channel is not None:
            return RemoteRepository(channel)

    @cachedproperty
    def path(self):
        return local(self._call_remote('get_path'))


class RemoteTransaction(RemoteCaller):
    def __enter__(self):
        #XXX: take remote commit into account
        return RevisionBuilderPath(None, '', self)
    def __exit__(self, etype,  eval, tb):
        if etype is None:
            self.commit()


class RemoteBackend(object):
    def __init__(self, backend, module, spec):
        self.spec = spec
        self.name = backend
        self.module_name = module
        self.gateway = makegateway(spec)
        channel = self.gateway.remote_exec("""
            from anyvc.remote.slave import start_controller
            start_controller(channel)
        """)
        channel.send(backend)
        channel.send(module)
        self._channel = channel.receive()
        if self._channel is None:
            raise ImportError('module %s not found on remote' % module)
        self._caller = RemoteCaller(self._channel)
        self.active = True

    def stop(self):
        #XXX: propperly shutdown the slave?
        self._channel.close()
        self.gateway.exit()
        self.active = False

    @property
    def features(self):
        return self._caller.features()

    def is_repository(self, path):
        return self._caller.is_repository(path=path)

    def is_workdir(self, path):
        return self._caller.is_workdir(path=path)

    def Repository(self, **kw):
        newchan = self._caller.open_repo(**kw)
        return RemoteRepository(newchan)

    def Workdir(self, path, **kw):
        kw['path'] = str(path)
        if 'source' in kw:
            kw['source'] = str(kw['source'])
        newchan = self._caller.open_workdir(**kw)
        return RemoteWorkdir(newchan)



repo_class = 'anyvc.git.repo:GitRepository'
workdir_class = 'anyvc.git.workdir:Git'

features = [
    'dvcs',
    'wd:heavy',
]


base_dirs = [
    'objects',
    'hooks',
    'info',
]

def has_git_base_dirs(path):
    return all(path.join(x).check(dir=1) for x in base_dirs)

def is_workdir(path):
    #XXX better check
    git = path.join('.git')
    return git.check(dir=1) and has_git_base_dirs(git)

def is_repository(path):
    return has_git_base_dirs(path) or has_git_base_dirs(path/'.git')



import py
from os import path
from anyvc.common.commit_builder import CommitBuilder


class FakeCommit(object):
    def __init__(self, file_tree):
        self.file_tree = file_tree
    
    def file_content(self, name):
        items = name.split(path.sep)
        tree = self.file_tree
        for item in items:
            tree = tree[item]
        return tree



def test_rename():
    base_commit = FakeCommit({
        'test':{
            'test.py':'#!/usr/bin/python',
            },
        })

    commit_builder = CommitBuilder(None, base_commit, author='test')

    commit_builder.mkdir('test2')
    commit_builder.rename('test/test.py', 'test2/test.py')

    content = commit_builder.read('test2/test.py')
    assert content == '#!/usr/bin/python'


    commit_builder.write('test2/test.py', 'test = 1')
    commit_builder.rename('test2/test.py', 'test/test.py')




# -*- coding: utf-8 -*- 
# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
"""
    Command line client for anyvc.

    :license: LGPL2 or later
    :copyright:
        * (c) 2008 Ali Afshar <aafshar@gmail.com>
        * (c) 2008 Ronny Pfannschmidt <Ronny.Pfannschmidt@gmx.de>
"""


import os, sys, logging

from optparse import OptionParser

# XXX Translations
_ = lambda s: s

from . import workdir


def create_option_parser():
    usage = "usage: %prog [options] <command> [args ...]"

    parser = OptionParser(usage)

    # General options
    parser.add_option('-d', '--working-directory', dest='working_directory',
                      help='The working directory')
    parser.add_option('-r', '--revision', dest='revision',
                      help='The revision id')

    # List options
    parser.add_option('-a', '--list-all', dest='list_all',
                      action='store_true', help='List all files.')
    parser.add_option('-m', '--message', dest='commit_message',
                     action='store', help='The commit message')
    parser.add_option('-U', '--hide-unknown', dest='hide_unknown',
                      action='store_true', help='Hide unknown files')
    parser.add_option('-u', '--list-unchanged', dest='list_unchanged',
                      action='store_true', help='List unchanged files.')
    parser.add_option('-i', '--list-ignored', dest='list_ignored',
                      action='store_true', help='List ignored files.')
    parser.add_option('-n', '--list-nonrecursive', dest='list_nonrecursive',
                      action='store_true',
                      help='Only list files in the current working directory.')
    parser.add_option('-c', '--no-color', dest='no_color',
                       action='store_true',
                       help='Uncoloured terminal list output')

    # Miscellaneous options
    parser.add_option('-v', '--verbose', action="store_true", dest="verbose",
                      help='Show debugging information')
    return parser


def setup_logger(verbose):
    if verbose:
        level=logging.DEBUG
    else:
        level=logging.INFO
    logging.basicConfig(level=level, stream=sys.stderr, format='%(levelname)s: %(message)s')


list_letters = {
    'clean': '-',
    'unknown': '?',
    'modified': 'M',
    'added': 'A',
    'removed': 'D',
    'deleted': '!',
    'ignored': 'I',
}

list_colors = {
    'clean': None,
    'unknown': '*teal*',
    'modified': 'red',
    'added': 'blue',
    'removed': 'fuscia',
    'deleted': 'yellow',
    'ignored': 'turquoise',
}


def do_status(vc, opts, args):

    try:
        from pygments import console
    except ImportError:
        console = None

    def output_state(st):
        output = list_letters.get(st.state, '*').ljust(2)
        color = list_colors.get(st.state)
        if (sys.stdout.isatty() 
            and console 
            and color 
            and not opts.no_color
            and os.environ.get('TERM') != "dumb"):
            output = console.ansiformat(color, output)
        sys.stdout.write(output)
        sys.stdout.write(st.relpath)
        sys.stdout.write('\n')
        sys.stdout.flush()

    hidden_states = []

    if not opts.list_all:

        if not opts.list_unchanged:
            hidden_states.append('clean')

        if not opts.list_ignored:
            hidden_states.append('ignored')

        if opts.hide_unknown:
            hidden_states.append('unknown')

    for st in vc.status(recursive=not opts.list_nonrecursive):
        if st.state not in hidden_states:
            output_state(st)

    return 0


def do_diff(vc, opts, args):
    try:
        from pygments import highlight
        from pygments.lexers import get_lexer_by_name
        from pygments.formatters import TerminalFormatter
        has_pygments = True
    except:
        has_pygments = False
        logging.debug('Pygments is not available.')

    paths = tuple(args)
    diff = vc.diff(paths=paths).strip()

    if (not opts.no_color 
        and has_pygments 
        and sys.stdout.isatty() 
        and os.environ.get('TERM') != 'dumb'):
        diff = highlight(diff, get_lexer_by_name('diff'), TerminalFormatter())

    sys.stdout.write(diff)
    sys.stdout.flush()

def do_commit(vc, opts, args):
    out = vc.commit(
        message=opts.commit_message,
        paths=args)
    sys.stdout.write(out)
    sys.stdout.flush()

def do_add(vc, opts, args):
    out = vc.add(paths=args)
    sys.stdout.write(out)
    sys.stdout.flush()


def do_push(vc, opts, args):
    repo = vc.repository
    if repo is None:
        print >>sys.stderr, "cant find local repo to push from"

    if not repo.local:
        #XXX: better handling
        print >>sys.stderr, "can't push from a non-local", repo.__class__.__name__
        exit(1)
    location = args[0] if args else None
    print repo.push(location, opts.revision)



# The available commands
#XXX: needs a better abstraction
commands = {
    'add': do_add,
    'status': do_status,
    'st': do_status,
    'diff': do_diff,
    'commit': do_commit,
    'ci': do_commit,
    'push': do_push,
}


def main(argv=sys.argv):
    parser = create_option_parser()
    opts, args = parser.parse_args(argv)

    setup_logger(opts.verbose)

    if opts.working_directory is None:
        cwd = os.getcwd()
    else:
        cwd = opts.working_directory

    logging.debug('Using working directory: %s' % cwd)

    vc = workdir.open(cwd)

    if vc is None:
        logging.error(_('Cannot detect version control system in %(cwd)s' %
                        {'cwd': cwd}))
        return -1

    logging.debug('Found VC: %s' % vc)

    pargs = args[:]
    called = pargs.pop(0)

    try:
        command = pargs.pop(0)
    except IndexError:
        logging.error(_('You must provide a command.'))
        logging.info(_('The available commands are: %(commands)s' %
                        {'commands': '%s' % ', '.join(commands.keys())}))
        parser.print_usage()
        return -1

    try:
        action = commands[command]
    except KeyError:
        logging.error(_('The command "%(command)s" is not available.' %
                        {'command': command}))
        logging.info(_('The available commands are: %(commands)s' %
                        {'commands': ', '.join(sorted(commands.keys()))}))
        return -1

    logging.debug('Calling %s' % command)

    return action(vc, opts, pargs)





def test_write_direct(repo):
    with repo.transaction(author='test', message='test') as root:
        root.join('test.txt').write('test')

def test_write_file(repo):
    with repo.transaction(author='test', message='test') as root:
        with root.join('test.txt').open('w') as file:
            file.write('test')




import py
from anyvc.metadata import backends, get_backend

def test_get_backend(backend, mgr):
    assert backend.module_name == backends[mgr.vc]


def test_has_features(backend):
    assert isinstance(backend.features, set)


def test_has_working_repository_check(repo, backend):
    print repo.path
    assert backend.is_repository(repo.path)


def test_has_working_workdir_check(wd, backend):
    print wd.path
    assert backend.is_workdir(wd.path)



"""
    Anyvc svn repo support

    :license: LGPL 2 or later
    :copyright: 2009 by Ronny Pfannschmidt
"""
import sys

import subvertpy
from subvertpy import repos, delta
from subvertpy.ra import RemoteAccess, Auth, get_username_provider, SubversionException
from anyvc.common.repository import Repository, Revision, join
from anyvc.common.commit_builder import CommitBuilder
from subvertpy.properties import time_from_cstring, time_to_cstring
from StringIO import StringIO
from ..exc import NotFoundError


from datetime import datetime


class SubversionRevision(Revision):
    def __init__(self, repo, id):
        #XXX: branch subdirs
         self.repo, self.id = repo, id

    @property
    def parents(self):
        #XXX: ignore irelevant revisions (no local change)
        #XXX: use mergeinfo to figure additional parents
        if self.id == 1: # rev 0 is the repo creation
            return []
        return [SubversionRevision(self.repo, self.id -1)]

    def file_content(self, path):
        try:
            target = StringIO()
            self.repo.ra.get_file(path.lstrip('/'), target, self.id)
            return target.getvalue()
        except: #XXX: bad bad
            raise IOError('%r not found'%path)


    def path_info(self, path):
        return self.repo.ra.check_path(path, self.id)

    def exists(self, path):
        kind = self.path_info(path.lstrip('/'))
        return kind in (subvertpy.NODE_FILE, subvertpy.NODE_DIR)

    def prop(self, name):
        return self.repo.ra.rev_proplist(self.id).get(name)

    @property
    def message(self):
        return self.prop('svn:log')

    def get_changed_files(self):
        files = []
        def callback(paths, rev, props, childs=None):
            #XXX: take branch path into account?
            files.extend(paths)
        self.repo.ra.get_log(
            start = self.id,
            end = self.id,
            callback = callback,
            paths = None,
            discover_changed_paths=True)
        return files

    @property
    def author(self):
        return self.prop('svn:author')

    @property
    def time(self):
        date_str = self.prop('svn:date')
        timestamp = time_from_cstring(date_str)
        #XXX: subertpy uses a magic factor of 1000000
        return datetime.fromtimestamp(float(timestamp)/1000000)


class SvnCommitBuilder(CommitBuilder):
    def commit(self):
        ra = RemoteAccess(self.repo.url,
                          auth=Auth([get_username_provider()]))
        editor = ra.get_commit_editor({
            'svn:log':self.extra['message'],

            #XXX: subertpy uses a magic factor of 1000000
            #XXX: subversion cant set a commit date on commit, sucker
            #'svn:date':time_to_cstring(self.time_unix*1000000),
            })
        print self.time
        root = editor.open_root()

        for src, target in self.renames:
            #XXX: directories
            src = src.lstrip('/')
            target = target.lstrip('/')
            file = root.add_file(target, join(self.repo.url, src), 1)
            file.close()
            root.delete_entry(src)

        for file in self.contents:
            try:
                svnfile = root.add_file(file)
            except:
                svnfile = root.open_file(file)
            txhandler = svnfile.apply_textdelta()
            f = StringIO(self.contents[file])
            delta.send_stream(f, txhandler)
            svnfile.close()
        root.close()
        editor.close()


class SubversionRepository(Repository):

    CommitBuilder = SvnCommitBuilder

    def __init__(self, path, create=False):
        #XXX: correct paths
        if create:
            repos.create(path.strpath)
        self.path = path
        self.url = "file://%s" % path
        self.ra = RemoteAccess(self.url)

    def __len__(self):
        return self.ra.get_latest_revnum()

    def get_default_head(self):
        #XXX: correct paths !!!, grab trunk
        last = len(self)
        if last == 0:
            return
        return SubversionRevision(self, last)

    def __getitem__(self, id):
        return SubversionRevision(self, id)




import sys
import os

from anyvc.common.workdir import WorkDirWithParser
from ..exc import NotFoundError

from StringIO import StringIO

#bzr imports
from bzrlib.branch import Branch
from bzrlib.workingtree import WorkingTree
from bzrlib.errors import NotBranchError
from bzrlib import bzrdir
from bzrlib import osutils
from bzrlib.status import show_tree_status
from bzrlib.diff import DiffTree
from bzrlib.diff import show_diff_trees
from bzrlib import revisionspec


class Bazaar(WorkDirWithParser):
    statemap  = {
            "unknown:": 'unknown',
            "added:": 'added',
            "unchanged:": 'clean',
            "removed:": 'removed',
            "ignored:": 'ignored',
            "modified:": 'modified',
            "conflicts:": 'conflict',
            "pending merges:": 'conflict',
            "renamed:": "placeholder", # special cased, needs parsing
            #XXX: figure why None didn't work
            }

    @property
    def repository(self):
        from .repo import BazaarRepository

        return BazaarRepository(workdir=self)

    def __init__(self, path, create=False, source=None):


        self.path = path

        if create:
            assert source
            source, _ = Branch.open_containing(str(source))
            source.bzrdir.sprout(path.strpath)

        try:
            self.wt = WorkingTree.open(path.strpath)
        except NotBranchError:
            raise NotFoundError(self.__class__, path)

    def status_impl(self, *k, **kw):
        #XXX: paths, recursion
        self.wt.lock_read()
        try:
            for change in self.wt.iter_changes(self.wt.basis_tree(),
                                               include_unchanged=True,
                                               want_unversioned=True,
                                               ):
                yield change
        finally:
            self.wt.unlock()

    def parse_status_item(self, change, cache):

        (file_id,
            paths, changed, versioned,
            parent, name, kind,
            executable) = change

        if file_id=='TREE_ROOT':
            return None

        #XXX: propperly handle removed vs deleted vs made untracked

        # paths -> renamed
        source, target = paths
        result_path = target if isinstance(target, unicode) else source

        # versioned add/remove
        old, new = versioned
        if new and not old:
            return 'added', result_path
        elif old and not new:
            # deleted ?!
            return 'removed', result_path
        elif not new and not old:
            return 'unknown', result_path
        elif source!=target:
            return None, paths
        elif changed:
            if kind[0] != kind[1]:
                return 'missing', result_path
            return 'modified', result_path
        elif all(versioned):
            return 'clean', result_path

        #XXX more tricky things ?!

    def add(self, paths=None, recursive=False):
        paths = self._abspaths(paths)
        try:
            added, ignored = self.wt.smart_add(paths,recursive)
            #XXX: more info?
        except:
            print "err"
            return "Error adding %s.\n%s" % (paths, sys.exc_value)
            #dialogs._bzrErrorDialog(_("Bazaar error adding file %s") % file,sys.exc_value)
        else:
            return "Ok"

    def commit(self, paths=None, message=None, user=None):
        paths = self._abspaths(paths)
        #XXX: this is weird
        if paths is not None:
            paths = map(self.wt.relpath, paths)
        self.wt.commit(message,author=user,specific_files=paths)

    def diff(self, paths=None):
        strdiff = StringIO()

        if paths is not None:
            paths = self._abspaths(paths)
            #XXX: this is weird
            paths = map(self.wt.relpath, paths)
        if paths is not None:
            show_diff_trees(self.wt.basis_tree(), self.wt, strdiff,
                            specific_files=paths)
        else:
            show_diff_trees(self.wt.basis_tree(), self.wt, strdiff)
        return strdiff.getvalue()

    def remove(self, paths=None, execute=False, recursive=False):
        assert paths is not None, 'uh wtf, dont do that till there is a sane ui'
        self.wt.remove(self._abspaths(paths), keep_files=False)

    def rename(self, source, target):
        #XXX: again the relpath weird :(
        self.wt.rename_one(
            self.wt.relpath(self.path.join(source).strpath),
            self.wt.relpath(self.path.join(target).strpath))

    def revert(self, paths=None, missing=False):

        revisionid = self.wt.branch.last_revision()
        ret = self.wt.revert(old_tree=self.wt.branch.repository.revision_tree(revisionid))

    def update(self, revision=None, paths=None):
        assert not revision and not paths
        #XXX fail
        self.wt.update()

    def _abspaths(self, paths):
        if paths is not None:
            return [ self.path.join(path).strpath for path in paths]

"""
To-Do
*'remove'
'rename'?
*'revert'
'status'?
*'update'
"""





"""
    anyvc.remote.object
    ~~~~~~~~~~~~~~~~~~~

    convience tool make execnet channels remote objects
"""
import pickle
from py.path import local

class RemoteCaller(object):
    def __init__(self, channel):
        self.channel = channel

    def _call_remote(self, name, *k, **kw):
        assert not self.channel.isclosed()

        if 'path' in kw:
            kw['path'] = str(kw['path'])

        data = (name, k, kw)
        self.channel.send(data)
        result = self.channel.receive()
        #XXX: again a job for py.execnet
        return result

    def __getattr__(self, name):
        def method(*k, **kw):
            return self._call_remote(name, *k, **kw)
        method.__name__ = name
        return method

class RemoteHandler(object):
    def __init__(self, channel):
        self.channel = channel
        channel.setcallback(self._channel_callback)

    def newchannel(self):
        return self.channel.gateway.newchannel()

    def _channel_callback(self, data):
        try:
            method, k, kw = data

            if 'path' in kw:
                kw['path'] = local(kw['path'])

            method = getattr(self, method)
            result = method(*k, **kw)
            self.channel.send(result)
        except Exception:
            #XXX: py.execnet should pass callback exceptions
            import traceback
            import sys
            excinfo = sys.exc_info()
            l = traceback.format_exception(*excinfo)
            errortext = "".join(l)
            self.channel.close(errortext)
        except:
            self.channel.close('unknown error')





repo_class = 'anyvc.mercurial.repo:MercurialRepository'
workdir_class = 'anyvc.mercurial.workdir:Mercurial'

features = [
    'dvcs',
    'wd:heavy',
]



def is_hg(path):
    return path.join('.hg/store').check(dir=1) \
       and path.join('.hg/requires').check()


is_repository = is_workdir = is_hg




#XXX: ValueError for backward compat, figure how to phase out later
class NotFoundError(LookupError, ValueError):
    """raised if a repo/workdir cant be found"""
    def __init__(self, vc, path):
        Exception.__init__(self, vc, path)
        self.vc = vc
        self.path = path

    def __str__(self):
        return "%(vc)s repo not found for %(path)r"%vars(self)



workdir_class = 'anyvc.subversion.workdir:SubVersion'
repo_class = 'anyvc.subversion.repo:SubversionRepository'

features = [
    'wd:light',
]


def is_workdir(path):
    svn = path.join('.svn')
    return svn.join('entries').check() \
       and svn.join('props').check(dir=1) \
       and svn.join('text-base').check(dir=1)

def is_repository(path):
    return path.join('format').check() \
       and path.join('hooks').check(dir=1) \
       and path.join('locks').check(dir=1) \
       and path.join('format').read().strip().isdigit()



"""
    this is a series of functional tests
    each begins at a given state and runs a series of actions,
    asserting the workdir state between each
"""


# copyright 2008 by Ronny Pfannschmidt
# license lgpl2 or later
import py
from anyvc.exc import NotFoundError
from anyvc.metadata import get_wd_impl


has_files = py.test.mark.files({
        'test.py':'print "test"',
        })

commited = py.test.mark.commit

@has_files
def test_workdir_add(wd):
    """
    * add a unknown file
    * commit it
    """

    wd.check_states(unknown=['test.py'])
    print wd.add(paths=['test.py'])
    wd.check_states(added=['test.py'])
    print wd.commit(paths=['test.py'], message='add test.py')
    wd.check_states(clean=['test.py'])

def test_subdir_state_add(wd):
    """
    * add a file in a subdir
    """
    wd.put_files({
        'subdir/test.py':'test',
    })

    print wd.add(paths=['subdir/test.py'])
    wd.check_states(added=['subdir/test.py'])
    print wd.commit(message='add subdir/test.py')
    wd.check_states(clean=['subdir/test.py'])


@has_files
@commited
def test_workdir_remove(wd):
    """
    * remove a known file
    * commit the removal
    """
    wd.check_states(clean=['test.py'])
    wd.remove(paths=['test.py'])
    wd.check_states(removed=['test.py'])
    wd.commit(message='*')

    py.test.raises(AssertionError,wd.check_states, clean=['test.py'])
    assert not wd.path.join('test.py').check()


@has_files
@commited
def test_workdir_rename(wd):
    """
    * rename a known file
    * check the add/remove pair
    * commit
    """
    wd.rename(source='test.py', target='test2.py')
    wd.check_states(
        removed=['test.py'],
        added=['test2.py'],
    )

    wd.commit(message='renamed')
    wd.check_states(clean=['test2.py'])


@has_files
@commited
def test_workdir_revert(wd):
    """
    * remove a known file (vcs command)
    * revert
    * change the content of a known file
    * revert
    """
    wd.remove(paths=['test.py'])
    wd.check_states(removed=['test.py'])

    wd.revert(paths=['test.py'])
    wd.check_states(clean=['test.py'])

    wd.put_files({
        'test.py':'oooo'
        })

    wd.check_states(modified=['test.py'])

    wd.revert(paths=['test.py'])
    wd.check_states(clean=['test.py'])


@has_files
def test_diff_all(wd):
    """
    * change a known file
    * check the diff
    """
    wd.add(paths=['test.py'])
    wd.commit(message='*')
    wd.put_files({
        'test.py':'ooo'
    })

    diff = wd.diff()
    print diff
    assert 'ooo' in diff
    assert 'print "test"' in diff


@has_files
@commited
def test_file_missing(wd):
    """
    * delete a known file (just in the workdir)
    * check the missing state
    """
    wd.delete_files('test.py')
    wd.check_states(missing=['test.py'])


@has_files
@commited
def test_status_subdir_only(wd):
    """
    * add a file in a subdir
    * commit
    * change the file in the subdir
    * check status for the subdir only returning items of that subdir
    """
    wd.put_files({
        'subdir/a.py':'foo\n',
        })
    wd.add(paths=['subdir/a.py'])
    wd.check_states(added=['subdir/a.py'])

    print wd.commit(message='add some subdir')

    wd.check_states(clean=['subdir/a.py'])
    wd.put_files({
        'subdir/a.py':'bar\nfoo\n', #XXX: different size needed for hg status
        })

    stats = list(wd.status(paths=['subdir']))
    assert any(s.relpath == 'subdir/a.py' for s in stats)

    wd.check_states(modified=['subdir/a.py'])


@has_files
@commited
def test_workdir_open(wd, backend):
    """
    check if anyvc.workdir.open works as expected
    """
    import anyvc
    wd2 = anyvc.workdir.open(wd.path)
    assert backend.is_workdir(wd2.path)




@has_files
@commited
def test_workdir_open_honors_ANYVC_IGNORED_WORKDIRS(monkeypatch, wd):
    """
    check that :func:`anyvc.workdir.open` honors `ANYVC_IGNORED_WORKDIRS`
    """
    import anyvc
    assert anyvc.workdir.open(wd.path) is not None
    monkeypatch.setenv('ANYVC_IGNORED_WORKDIRS', wd.path)
    assert anyvc.workdir.open(wd.path) is None


def test_disallowed_paths(monkeypatch):
    from anyvc._workdir import _disallowd_workdirs
    monkeypatch.setenv('ANYVC_IGNORED_WORKDIRS', '/a:/b')
    dirs = _disallowd_workdirs()
    assert '/a' in dirs
    assert '/b' in dirs
    monkeypatch.setenv('ANYVC_IGNORED_WORKDIRS', '~')
    dirs = _disallowd_workdirs()
    assert py.std.os.environ['HOME'] in dirs




"""
    Anyvc bzr repo support

    :license: LGPL 2 or later
    :copyright: 2009 by Ronny Pfannschmidt
"""
from bzrlib.bzrdir import BzrDir
from bzrlib.branch import Branch
from bzrlib import errors
from bzrlib.memorytree import MemoryTree
from anyvc.common.repository import Repository, Revision, join
from anyvc.common.commit_builder import CommitBuilder
from datetime import datetime
from ..exc import NotFoundError

class BazaarRevision(Revision):
    def __init__(self, repo, bzrrev):
        self.repo, self.bzrrev = repo, bzrrev

    @property
    def id(self):
        return self.bzrrev.revision_id

    def get_changed_files(self):
        # TODO: this doesn't yet handle the case of multiple parent revisions
        #XXX: handle first commit beter
        rev = self.bzrrev
        repo = self.repo.branch.repository

        current = repo.revision_tree(rev.revision_id)
        if self.parents:
            prev = repo.revision_tree(rev.parent_ids[0])
        else:
            prev = repo.revision_tree('null:')
        delta = current.changes_from(prev)
        files = [f[0] for f in delta.added + delta.removed + delta.renamed + delta.kind_changed + delta.modified]
        return files


        # diff_file is a stringio
        #diff_tree = diff.DiffTree(prev, current, diff_file)

        #self._branch.lock_read()
        #diff_tree.show_diff('')
        #self._branch.unlock()



    @property
    def time(self):
        return datetime.fromtimestamp(self.bzrrev.timestamp)

    @property
    def parents(self):
        return [ BazaarRevision(self.repo,
                                self.repo.branch.repository.get_revision(rev))
                for rev in self.bzrrev.parent_ids]
    @property
    def message(self):
        return self.bzrrev.message

    @property
    def author(self):
        return self.bzrrev.get_apparent_author()

    def file_content(self, path):
        tree = self.repo.branch.repository.revision_tree(self.bzrrev.revision_id)
        id = tree.path2id(path)
        if id is None:
            raise IOError('%r not found'%path)
        try:
            tree.lock_read()
            sio = tree.get_file(id)
            return sio.read()
        finally:
            tree.unlock()

    def exists(self, path):
        tree = self.repo.branch.repository.revision_tree(self.bzrrev.revision_id)
        return tree.path2id(path) is not None



class BazaarCommitBuilder(CommitBuilder):
    def __init__(self, *k, **kw):
        super(BazaarCommitBuilder, self).__init__(*k, **kw)
        self.tree = MemoryTree.create_on_branch(self.repo.branch)

    def __enter__(self):
        self.tree.lock_write()
        if self.tree.path2id('') is None:
            self.tree.add([''])
        return super(BazaarCommitBuilder, self).__enter__()

    def commit(self):
        tree = self.tree

        for name, content in self.contents.items():
            if not tree.path2id(name):
                tree.add([name])
            id = tree.path2id(name)
            tree.put_file_bytes_non_atomic(id, content)

        for old, new in self.renames:
            print old, '->', new
            tree.rename_one(old.lstrip('/'), new.lstrip('/'))


        self.tree.commit(
                message=self.extra['message'],
                authors=[self.author],
                timestamp=self.time_unix,
                timezone=self.time_offset,
                )

    def mkdir(self, path):
        #XXX: sanity checks?
        self.tree.mkdir(path)

    def __exit__(self, et, ev, tb):
        super(BazaarCommitBuilder, self).__exit__(et, ev, tb)
        self.tree.unlock()
        self.tree = None


class BazaarRepository(Repository):
    CommitBuilder = BazaarCommitBuilder

    #XXX: this whole thing is broken and messed
    def __init__(self, path=None, workdir=None, create=False):
        self.path = path or workdir.path
        if workdir:
            assert not path and not create
            self.branch = workdir.wt.branch
        elif create:
            assert path, 'create needs a path'
            self.branch = BzrDir.create_branch_convenience(str(path))
        else:
            try:
                self.branch, rest = Branch.open_containing(str(path))
            except errors.NotBranchError:
                raise NotFoundError('bzr', path)


    def __repr__(self):
        return "<Bzr 'repo' at %s>"%self.branch.base

    def __len__(self):
        repo = self.branch.repository
        stats = repo.gather_stats()
        return stats['revisions']

    def get_default_head(self):
        id = self.branch.last_revision()
        if id == "null:":
            return
        return self[id]

    def __getitem__(self, id):
        revision = self.branch.repository.get_revision(id)
        return BazaarRevision(self, revision)



    def push(self, *k, **kw):
        parent = self.branch.get_parent()
        remote = Branch.open(parent)

        self.branch.push(remote)




from anyvc.util import cachedproperty
from itertools import count

def test_cached_property():
    # kinda stupid
    class test(object):
        def __init__(self):
            self.count = count()

        @cachedproperty
        def prop(self):
            return self.count.next()

    t = test()

    assert t.prop == 0
    assert t.prop == 0
    del t.prop
    assert t.prop == 1



workdir_class = 'anyvc.bazaar.workdir:Bazaar'
repo_class = 'anyvc.bazaar.repo:BazaarRepository'

features = [
#XXX: make wd without that?
#    'dvcs',
]

def is_workdir(path):
    return path.join('.bzr/checkout').check(dir=1)

def is_repository(path):
    return path.join('.bzr/repository').check(dir=1)



"""
    monotone backend
    ----------------

    :copyright: 2007-2008 Ronny Pfannschmidt
    :license: LGPL2 or later
"""

from cmdbased import CommandBased, relative_to, Path
import os.path

class Monotone(CommandBased):

    cmd = 'mtn'
    detect_subdir = '_MTN'

    def __init__(self, versioned_path):
        CommandBased.__init__(self, versioned_path)
        self.interface_version = float(
                self.execute_command(['au','interface_version']))

    def process_paths(self, paths):
        return map(relative_to(self.base_path), paths)

    def get_remove_args(self, paths, **kw):
        return ['drop'] + self.process_paths(paths)

    def get_list_args(self, paths=(), recursive=True, **kw):
        cmd = ['automate', 'inventory']
        if self.interface_version >= 6:
            # we got a recent version with better features
            if not recursive:
                cmd.extend(['--depth', '0'])
            cmd.extend(self.process_paths(paths))
        return cmd

    def parse_list_item(self, item, cache):
        state = self.statemap.get(item[:3], 'none')
        return Path(os.path.normpath(item[8:].rstrip()), state, self.base_path)

    def parse_list_items(self, items, cache):
        if self.interface_version < 6:
            return (self.parse_list_item(item, cache) for item in items)
        else:
            return self.parse_basic_io(items)

    def parse_basic_io(self, iter):
        """
        .. warning::
            * this doesnt parse correct, just the minimal work to get the status info
            * no support for filenames with spaces 
            * no complete status handling yet
        """
        for info in self.parse_basic_io_dicts(iter):
            stats = info['status']
            path = info['path'][0]
            changes = info.get('changes', [])

            if not path or path == '.':
                continue

            #XXX: REALLY RECHECK THIS !!!!
            # print path, stats, changes

            if 'rename_source' in stats and not 'rename_target' in stats:
                continue

            if 'invalid' in stats:
                status = 'error'
            elif 'content' in changes or 'rename_target' in changes:
                status = 'modified'
            elif 'known' in stats:
                status = 'normal'
            elif 'unknown' in stats:
                if 'dropped' in stats:
                    status = 'removed'
                else:
                    status = 'unknown'
            elif 'ignored' in stats:
                status = 'ignored'
            elif 'missing' in stats:
                status = 'missing'
            else:
                status = 'error'


            yield Path(path, status, self.base_path)

    def parse_basic_io_dicts(self, iter):
        res = {}
        for line in iter:
            line = line.strip().split()
            if line:
                res[line[0]] = [x.strip('"') for x in line[1:]]
            else:
                yield res
                res = {}
    
    # for monotone prior to 0.37
    statemap = {
        '   ': 'normal',   # unchanged
        '  P': 'modified', # patched (contents changed)
        '  U': 'none',     # unknown (exists on the filesystem but not tracked)
        '  I': 'ignored',  # ignored (exists on the filesystem but excluded by lua hook)
        '  M': 'missing',  # missing (exists in the manifest but not on the filesystem)

        ' A ': 'error',    # added (invalid, add should have associated patch)
        ' AP': 'new',      # added and patched
        ' AU': 'error',    # added but unknown (invalid)
        ' AI': 'error',    # added but ignored (seems invalid, but may be possible)
        ' AM': 'empty',    # added but missing from the filesystem

        ' R ': 'normal',   # rename target
        ' RP': 'modified', # rename target and patched
        ' RU': 'error',    # rename target but unknown (invalid)
        ' RI': 'error',    # rename target but ignored (seems invalid, but may be possible?)
        ' RM': 'missing',  # rename target but missing from the filesystem

        'D  ': 'removed',  # dropped
        'D P': 'error',    # dropped and patched (invalid)
        'D U': 'error',  # dropped and unknown (still exists on the filesystem)
        'D I': 'error',    # dropped and ignored (seems invalid, but may be possible?)
        'D M': 'error',    # dropped and missing (invalid)

        'DA ': 'error',    # dropped and added (invalid, add should have associated patch)
        'DAP': 'new',      # dropped and added and patched
        'DAU': 'error',    # dropped and added but unknown (invalid)
        'DAI': 'error',    # dropped and added but ignored (seems invalid, but may be possible?)
        'DAM': 'missing',  # dropped and added but missing from the filesystem

        'DR ': 'normal',   # dropped and rename target
        'DRP': 'modified', # dropped and rename target and patched
        'DRU': 'error',    # dropped and rename target but unknown (invalid)
        'DRI': 'error',    # dropped and rename target but ignored (invalid)
        'DRM': 'missing',  # dropped and rename target but missing from the filesystem

        'R  ': 'missing',  # rename source
        'R P': 'error',    # rename source and patched (invalid)
        'R U': 'removed',  # rename source and unknown (still exists on the filesystem)
        'R I': 'error',    # rename source and ignored (seems invalid, but may be possible?)
        'R M': 'error',    # rename source and missing (invalid)

        'RA ': 'error',    # rename source and added (invalid, add should have associated patch)
        'RAP': 'new',      # rename source and added and patched
        'RAU': 'error',    # rename source and added but unknown (invalid)
        'RAI': 'error',    # rename source and added but ignored (seems invalid, but may be possible?)
        'RAM': 'missing',  # rename source and added but missing from the filesystem

        'RR ': 'new',      # rename source and target
        'RRP': 'modified', # rename source and target and target patched
        'RRU': 'error',    # rename source and target and target unknown (invalid)
        'RRI': 'error',    # rename source and target and target ignored (seems invalid, but may be possible?)
        'RRM': 'missing',   # rename source and target and target missing
    }




"""
    Anyvc git repo support

    :license: LGPL 2 or later
    :copyright: 2009 by Ronny Pfannschmidt
"""


from anyvc.common.repository import Repository, Revision
from anyvc.common.commit_builder import CommitBuilder
from ..exc import NotFoundError
import subprocess
import stat
import os
from py.path import local
from datetime import datetime
from collections import defaultdict
from dulwich.repo import Repo
from dulwich.objects import Blob, Tree, Commit
from dulwich.errors import NotGitRepository

class GitRevision(Revision):

    def __init__(self, repo, commit):
        self.repo, self.commit = repo, commit

    @property
    def id(self):
        return self.commit.id

    @property
    def author(self):
        return self.commit.author

    @property
    def time(self):
        #XXX distinct author and commiters?
        return datetime.fromtimestamp(self.commit.commit_time)

    @property
    def parents(self):
        return [GitRevision(self.repo, self.repo.repo[id])
                for id in self.commit.parents]

    @property
    def message(self):
        return self.commit.message.rstrip()


    def resolve(self, path):
        repo = self.repo.repo
        tree = repo[self.commit.tree]
        loc, name = os.path.split(path.lstrip('/'))
        for part in loc.split(os.sep):
            if not part:
                continue
            tree = repo[tree[part][1]]
        return repo[tree[name][1]]

    def file_content(self, path):
        try:
            #XXX: highly incorrect, should walk and check the type
            blob = self.resolve(path)
            return blob.data
        except KeyError:
            raise IOError('%r not found'%path)

    def exists(self, path):
        try:
            self.resolve(path)
            return True
        except KeyError:
            return False

    def get_changed_files(self):
        new = self.commit.tree
        if self.parents:
            old = self.parents[0].commit.tree
        else:
            old = None
        #XXX: bad code
        added, removed, changed = diff_tree(self.repo.repo, old, new)
        return sorted(added | removed | changed)

class GitCommitBuilder(CommitBuilder):

    def commit(self):
        #XXX: evidence for the rest of 
        # this functions is supposed not to exist
        # yes, its that 

        #XXX: generate all objects at once and 
        #     add them as pack instead of legacy objects

        r = self.repo.repo
        store = r.object_store
        new_objects = []
        names = sorted(self.contents)
        nametree = defaultdict(list)
        for name in names:
            base = name.strip('/')
            while base:
                nbase = os.path.dirname(base)
                nametree[nbase].append(base)
                base = nbase

        if self.base_commit:
            tree = r.tree(self.base_commit.commit.tree)
            tree._ensure_parsed()
            print tree._entries
        else:
            tree = Tree()

        for src, dest in self.renames:
            src = src.strip('/')
            dest = dest.strip('/')
            tree[dest] = tree[src]
            del tree[src]


        for name in names:
            blob = Blob()
            blob.data = self.contents[name]
            new_objects.append((blob, name))
            tree.add(0555, os.path.basename(name), blob.id)
        new_objects.append((tree, ''))
        commit = Commit()
        if self.base_commit:
            commit.parents = [self.base_commit.commit.id]
        commit.tree = tree.id
        commit.message = self.extra['message']
        commit.committer = self.author
        commit.commit_time = int(self.time_unix)
        commit.commit_timezone = self.time_offset
        commit.author = self.author
        commit.author_time = int(self.time_unix)
        commit.author_timezone = self.time_offset
        new_objects.append((commit, ''))
        store.add_objects(new_objects)
        self.repo.repo.refs['HEAD'] = commit.id

class GitRepository(Repository):
    CommitBuilder = GitCommitBuilder

    def __init__(self, path=None, workdir=None, create=False, bare=False):
        assert path or workdir
        if workdir:
            self.path = workdir.path
        else:
            self.path = path
        if create:
            #XXX: fragile
            path.ensure(dir=1)
            self.repo = Repo.init(path.strpath)
        else:
            assert self.path.check(dir=True)
            try:
                self.repo = Repo(self.path.strpath)
            except NotGitRepository:
                raise NotFoundError('git', self.path)

    def __len__(self):
        #XXX: fragile
        head = self.get_default_head()
        if head is None:
            return 0
        return len(self.repo.revision_history(head.id))

    def push(self):
        #XXX: hell, figure if the remote is empty, push master in that case
        #XXX: use dulwich?
        subprocess.check_call(['git', 'push', '--all'], cwd=self.path)

    def get_default_head(self):
        revs = self.repo.get_refs()
        head = revs.get('HEAD', revs.get('master'))
        if head is not None:
            return GitRevision(self, self.repo.get_object(head))

    def __getitem__(self, id):
        return GitRevision(self, self.repo.get_object(id))




def walk_tree_object(repo, tree, path=''):
    for mode, name, sha in tree.entries():
        if stat.S_IFDIR & mode:
            for name, sha in walk_tree(repo, sha, name):
                yield os.path.join(path, name), sha
        else:
            yield os.path.join(path, name), sha

def walk_tree(repo, tree_id, path=''):
    if tree_id is None:
        return ()
    tree = repo.get_object(tree_id)
    assert tree.__class__ is Tree
    return walk_tree_object(repo, tree, path)


def diff_tree(repo, old, new):
    #XXX: bad code
    new_sha = dict(walk_tree(repo, new))
    old_sha = dict(walk_tree(repo, old))
    new_set = set(new_sha)
    old_set = set(old_sha)

    added = new_set - old_set
    removed = old_set - new_set

    changed = set(name
            for name, sha in new_sha.items()
            if name not in added
            and name not in removed
            and old_sha[name] != sha)
    return added, removed, changed








"""
    anyvc.util
    ~~~~~~~~~~

    some utility classes that help with metadata
"""

class cachedproperty(object):
    def __init__(self, func):
        self.func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

    def __get__(self, obj, type):
        if obj is None:
            return self
        else:
            result = self.func(obj)
            setattr(obj, self.__name__, result)
            return result




# -*- coding: utf-8 -*-
# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
"""
    anyvc.mercurial
    ~~~~~~~~~~~~~~~

    Mercurial support

    :license: LGPL2 or later
    :copyright: 2008 Ronny Pfannschmidt
"""
import os
from functools import wraps

from anyvc.common.workdir import WorkDir, StatedPath
from ..exc import NotFoundError

from mercurial.__version__ import version as hgversion
# no support for hg <= 1.0.2
if hgversion < '1.2':
    raise ImportError('HG version too old, please update to a release >= 1.2')

from mercurial import ui as hgui, hg, commands, util, cmdutil
from mercurial.match import match, always
try:
    remoteui = cmdutil.remoteui
except AttributeError:
    remoteui = hg.remoteui

__all__ = 'Mercurial',



#XXX: this shouldn't work if used by the vc client
#     console output should be responsive
def grab_output(func):
    """
    wraps a call to hg and grabs the output
    """
    @wraps(func)
    def grabber(self, *k, **kw):
        self.repo.ui.pushbuffer()
        try:
            func(self, *k, **kw)
            return self.repo.ui.popbuffer()
        except Exception, e:
            e.hg_output = self.repo.ui.popbuffer()
            raise

    return grabber


class Mercurial(WorkDir):

    @property
    def repository(self):
        from .repo import MercurialRepository
        return MercurialRepository(workdir=self)

    def __init__(self, path, create=False, source=None):
        """
        Get a repo for a given path.
        If `create` is true, a new repo is created.
        """
        try:
            self.ui = hgui.ui(interactive=False, verbose=True, debug=True)
        except TypeError: # hg >= 1.3 ui
            self.ui = hgui.ui()
            self.ui.setconfig('ui', 'interactive', 'off')

        super(Mercurial, self).__init__(path, create, source)
        self.repo = hg.repository(self.ui, path.strpath)
        self.ui = self.repo.ui

    def create(self):
        hg.repository(self.ui, self.path.strpath, create=True)

    def setup(self):
        pass #XXX unneded

    def create_from(self, source):
        hg.clone(remoteui(self.ui, {}),
                str(source), self.path.strpath)

    def status(self, paths=(), *k, **kw):
        glob = '**' if kw.get('recursive') else '*'
        #XXX: merce conflicts ?!
        names = (
                'modified', 'added', 'removed',
                'missing', 'unknown', 'ignored', 'clean',
                )

        if paths:
        #XXX: investigate cwd arg
            matcher = match(self.repo.root, self.repo.root, paths,
                    default='relpath')
        else:
            matcher = always(self.repo.root, self.repo.root)

        state_files = self.repo.status(
            match=matcher,
            ignored=True,
            unknown=True,
            clean=True,
        )
        for state, files in zip(names, state_files):
            for file in files:
                yield StatedPath(file, state, base=self.path)



    @grab_output
    def add(self, paths=()):
        commands.add(self.ui, self.repo, *self.joined(paths))

    def joined(self, paths):
        if paths is None:
            return []
        return [os.path.join(self.repo.root, path) for path in paths]

    @grab_output
    def commit(self, paths=(), message=None, user=None):
        commands.commit(self.ui, self.repo,
            user=user,
            message=message,
            logfile=None,
            date=None,

            *self.joined(paths)
            )

    @grab_output
    def remove(self, paths):
        #XXX: support for after ?
        commands.remove(self.ui, self.repo,
                after=False, # only hg 0.9.5 needs that explicit
                force=False,
                *self.joined(paths)
                )

    @grab_output
    def revert(self, paths, rev=None): #XXX: how to pass opts['all']?
        if rev is None:
            parents = self.repo.parents()
            if len(parents)!=1 and rev is None:
                #XXX: better exception type?
                raise Exception(
                        "can't revert on top of a merge without explicit rev")
            rev = parents[0].rev()
        commands.revert(self.ui, self.repo,
                date=None,
                rev=rev,
                no_backup=False,
                *self.joined(paths))

    @grab_output
    def rename(self, source, target):
        commands.rename(self.ui, self.repo,
                after=False, # hg 0.9.5
                *self.joined([source, target])
                )

    @grab_output
    def diff(self, paths=(), rev=None):
        commands.diff(
                self.ui,
                self.repo,
                rev=rev,
                git=True,
                *self.joined(paths))








import py

@py.test.mark.xfail
@py.test.mark.incomplete
def test_simple_rename(mgr):
    repo = mgr.make_repo('repo')
    with repo.transaction(
            message='initial conent',
            author='test',
            ) as root:
        root.join('test.py').write('test = 1\n')
    with repo.transaction(
            message='add a rename',
            author='test',
            ) as root:
        assert root.join('test.py').read() == 'test = 1\n'

        root.join('test.py').rename('test2.py')

        assert root.join('test2.py').read() == 'test = 1\n'
        assert not root.join('test.py').check()





"""
    Anyvc Backend loading abstraction

    '

    :license: LGPL2
    :copyright: 2009 by Ronny Pfannschmidt
"""

from anyvc.util import cachedproperty
from anyvc.common.repository import Repository
from anyvc.common.workdir import WorkDir, WorkDirWithParser, CommandBased

class Backend(object):
    def __init__(self, name, module_name):
        self.name = name
        self.module_name = module_name

    def __repr__(self):
        return '<anyvc backend %s>'%(self.name,)

    def _import(self, name):
        module, attr = name.split(':')
        impl_module = __import__(module, fromlist=['*'])
        return getattr(impl_module, attr)

    def is_workdir(self, path):
        return self.module.is_workdir(path)

    def is_repository(self, path):
        return self.module.is_repository(path)

    @cachedproperty
    def module(self):
        return __import__(self.module_name, fromlist=['*'])

    @cachedproperty
    def features(self):
        return set(self.module.features)

    @cachedproperty
    def Repository(self):
        return self._import(self.module.repo_class)

    @cachedproperty
    def Workdir(self):
        return self._import(self.module.workdir_class)



"""
    Anyvc Repository Base Classes
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Various base classes for dealing with history data

    :license: LGPl 2 or later
    :copyright:
        * 2008 by Ronny Pfannschmidt <Ronny.Pfannschmidt@gmx.de>

    .. warning::

        the repo apis are unstable and incomplete

"""
from collections import defaultdict
from os.path import join, dirname
try:
    from io import BytesIO as StringIO #XXX: epic mess
except ImportError:
    from StringIO import StringIO


class MemoryFile(StringIO):
    def __init__(self, data='', path=None):
        StringIO.__init__(self, data)
        self.path = path

    def __enter__(self):
        return self

    def __exit__(self, et, ev, tb):
        pass

class Revision(object):

    def get_parent_diff(self): 
        from anyvc.diff import diff_for_commit
        return diff_for_commit(self)

    def __enter__(self): 
       return RevisionView(self, '/')

    def __exit__(self, et, ev , tb):
        pass


class RevisionView(object):
    def __init__(self, revision, path):
        self.revision = revision
        self.path = path

    def join(self, path):
        return RevisionView(self.revision, join(self.path, path))

    def read(self):
        return self.revision.file_content(self.path)

    def open(self):
        return MemoryFile(self.read(), self.path)

    def isdir(self):
        self.revision.isdir(self.path)

    def isfile(self):
        return self.revision.isfile(self.path)

    def exists(self):
        return self.revision.exists(self.path)



class Repository(object):
    """
    represents a repository
    """

    local = True

    def __init__(self,**extra):
        self.path = path
        self.extra = extra

    def prepare_default_structure(self):
        """
        if the vcs has a common standard repo structure, set it up
        """
        pass

    def push(self, dest=None, rev=None):
        """
        push to a location

        :param dest: the destination
        :param rev: the maximum revision to push, may be none for latest
        """
        raise NotImplementedError("%r doesnt implement push"%self.__class__)

    def pull(self, source=None, rev=None):
        """
        counterpart to push
        """
        raise NotImplementedError("%r doesnt implement pull"%self.__class__)


    def transaction(self, **extra):
        # will be set by subclasses
        return self.CommitBuilder(self, self.get_default_head(), **extra)





#!/usr/bin/python
from __future__ import with_statement
import os, sys
from commands import getoutput
try:
    from setuptools import setup
except:
    from distutils.core import setup
import time
import hgdistver

def read_readme():
    with open('docs/readme.rst') as f:
        return f.read()

setup(
    version = hgdistver.get_version(cachefile='anyvc/__version__.py'),
    name = 'anyvc',
    packages = [
        'anyvc',
        'anyvc.common',
        'anyvc.remote',

        # backends
        'anyvc.mercurial',
        'anyvc.git',
        'anyvc.subversion',
        'anyvc.bazaar',
    ],
    install_requires=[
        'apipkg',
        'py>=1.3',
    ],
    extras_require={
        'mercurial': ['mercurial'],
        'bazaar': ['bzr'],
        'git': ['dulwich'],
        'subversion': ['subvertpy'],
        'remoting': ['execnet'],
    },

    scripts=[
        'bin/vc',
    ],
    description='Library to access any version control system.',
    license='GNU GPL2 (or later) as published by the FSF',
    url='http://www.bitbucket.org/RonnyPfannschmidt/anyvc/',
    author='Ronny Pfannschmidt',
    author_email='Ronny.Pfannschmidt@gmx.de',
    long_description=read_readme(),
    classifiers = [
        'Intended Audience :: Developers',
    ],
)



"""
    anyvc diff tools

    :license: LGPL2 or later
"""
import itertools
import difflib

#XXX: those are a pile of hacks

def diff_for_file(commit, file):
    #XXX: propperly handle add/remove
    try:
        new = commit.file_content(file).splitlines(True)
    except: #XXX: ERRORS?!?!
        new = []

    if commit.parents:
        parent = commit.parents[0]
        try:
            old = parent.file_content(file).splitlines(True)
        except: #XXX: ERRORS?!?!
            old = []

    else:
        old = []

    #XXX: propperly handle lack of tailing \n
    #XXX: binaries ?!
    return difflib.unified_diff(old, new, file, file)

def diff_for_commit(commit):
    changed_files = commit.get_changed_files()

    return ''.join(
            itertools.chain.from_iterable(
                diff_for_file(commit, file)
                for file in changed_files
            ))



class Darcs(CommandBased):
    #TODO: ensure this really works in most cases
    #XXX: path processing?

    cmd = 'darcs'
    detect_subdir = '_darcs'

    def get_commit_args(self, message, paths=(), **kw):
        return ['record', '-a', '-m', message] + list(paths)

    def get_revert_args(self, paths=()):
        return ['revert', '-a'] + list(paths)

    def get_status_args(self, **kw):
        return ['whatsnew', '--boring', '--look-for-adds']

    state_map = {
        "a": 'unknown',
        "A": 'added',
        "M": 'modified',
        "C": 'conflict',
        "R": 'removed'
    }

    move_regex = re.compile(" (?P<removed>.*?) -> (?P<added>.*?)$")


    def parse_status_item(self, item, cache):
        if item.startswith('What') or item.startswith('No') or not item.strip():
            return
        match = self.move_regex.match(item)
        if match:
            return None, (match.group('removed'),  match.group('added'))

        elements = item.split(None, 2)[:2] #TODO: handle filenames with spaces
        state = self.state_map[elements[0]]
        file = os.path.normpath(elements[1])
        return state, file

    def get_rename_args(self, source, target):
        return ['mv', source, target]






# copyright 2008 by Ronny Pfannschmidt
# license lgpl3 or later

from __future__ import with_statement

from anyvc.metadata import state_descriptions

class WdWrap(object):
    """
    :param wd: the workdir to wrap
    :type wd: subclass of :class:`anyvc.common.workdir.Workdir`

    decorator for a vcs workdir instance
    adds testing utility functions and proxies the other methods/attributes
    to the real instance
    """
    def __init__(self, wd):
        self.__wd = wd

    def __getattr__(self, name):
        return getattr(self.__wd, name)


    def put_files(self, mapping):
        """
        :type mapping: dict of (filename, text content)

        the text content will be rstripped and get a newline appended
        """
        for name, content in mapping.items():
            path = self.path.ensure(name)
            path.write(content.rstrip() + '\n')

    def has_files(self, *files):
        """
        :arg files: a listing of filenames that shsould exist
        """
        missing = [name for name in map(self.path.join, files) if not name.check()]
        assert not missing, 'missing %s'%', '.join(missing)
        return not missing

    def delete_files(self, *relpaths):
        """
        :arg relpaths: listing of files to remove
        """
        for path in relpaths:
            self.path.join(path).remove()

    def check_states(self, exact=True, **kw):
        """
        .. better listing of the states!

        :param bool exact: if true, ignore additional states
        :keyword $statename: state name for that particular file list
        :type $statename: list of relative path
        :returns: True if all supplied files have the asumed state

        """
        __tracebackhide__ = True
        assert isinstance(exact, bool)
        mapping = dict((rn, state) for state, rns in kw.items() for rn in rns)
        print mapping
        used = set()
        all = set()
        infos = list(self.status())
        print infos
        for info in infos:
            all.add(info.abspath)
            print info
            assert info.state in state_descriptions
            if info.relpath in mapping:
                expected = mapping[info.relpath]
                assert info.state==expected, "%s wanted %s but got %s"%(
                        info.relpath,
                        expected,
                        info.state,
                        )
                used.add(info.relpath)

        untested = set(mapping) - used

        print 'all:', sorted(all)
        print 'used:', sorted(used)
        print 'missing?:', sorted(all - used)
        print 'untested:', sorted(untested)
        assert not untested , 'not all excepted stated occured'


class VcsMan(object):
    """
    utility class to manage the creation of repositories and workdirs
    inside of a specific path (usually the tmpdir funcarg of a test)
    """
    def __init__(self, vc, base, xspec, backend):
        self.remote = xspec is not None
        self.vc = vc
        self.base = base.ensure(dir=True)
        self.xspec = xspec
        self.backend = backend

    def __repr__(self): 
        return '<VcsMan %(vc)s %(base)r>'%vars(self)

    def create_wd(self, workdir, source=None):
        """
        :param workdir: name of the target workdir
        :type workdir: str
        :param source: name of a source repository
        :type source: repo or None

        create a workdir if `source` is given, use it as base
        """
        path = self.base/workdir
        source_path = getattr(source, 'path', None)
        wd = self.backend.Workdir(path, create=True, source=source_path)
        return WdWrap(wd)

    def make_repo(self, name):
        """
        :param name: name of the repository to create

        create a repository using the given name
        """
        return self.backend.Repository(path=self.base/name, create=True)




import py
import anyvc

@py.test.mark.files({'setup.py': 'pass'})
@py.test.mark.commit
@py.test.mark.feature('wd:light')
def test_checkout_local(repo, wd, mgr):
    path = mgr.base/'checkout'

    wd2 = anyvc.workdir.checkout(
        target=path,
        source=repo.path,
        )


@py.test.mark.files({'setup.py': 'pass'})
@py.test.mark.commit
@py.test.mark.feature('wd:heavy')
def test_clone(wd, mgr):

    path = mgr.base/'clone'

    wd2 = anyvc.workdir.clone(
        target=path,
        source=wd.path,
        )





# -*- coding: utf-8 -*- 
# vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:
"""
    anyvc
    ~~~~~

    Simple vcs abstraction.

    :license: LGPL2 or later
    :copyright:
        * 2006 Ali Afshar aafshar@gmail.com
        * 2008 Ronny Pfannschmidt Ronny.Pfannschmidt@gmx.de
"""
__all__ = ["all_known", "get_workdir_manager_for_path"]

import warnings

from .metadata import get_backends
import os
from py.path import local
from .common.workdir import find_basepath


def _disallowd_workdirs():
    return set(
        local(os.path.expanduser(x))
        for x in os.environ.get('ANYVC_IGNORED_WORKDIRS',
                                '/').split(os.pathsep))

def open(path):
    """
    :param path:
        a local path to the worktree
        preferable a `py.path.local` instance

    open a scm workdir

    It uses the backend metadata to find the correct backend and
    won't import unnecessary backends to keep the import time low
    """
    dont_try = _disallowd_workdirs()
    for part in local(path).parts(reverse=True):
        if part in dont_try:
            continue

        applying = [ backend for backend in get_backends()
                     if backend.is_workdir(part) ]

        if applying:
            if len(applying) > 1:
                warnings.warn('found more than one backend below %s' % part)
            return applying[0].Workdir(part)



def clone(source, target):
    """
    create a heavy checkout/clone of the given source
    """
    #XXX: remote support
    for backend in get_backends():
        if 'wd:heavy' in backend.features and backend.is_repository(source):
            return backend.Workdir(target, create=True, source=source)


def checkout(source, target):
    """
    create a light checkout of the given source
    """
    #XXX: remote support
    for backend in get_backends():
        if 'wd:light' in backend.features and backend.is_repository(source):
            #XXX: there should be an actual argument
            #     to keep heavy and light apart
            return backend.Workdir(target, create=True, source=source)
